from typing import Union
import pytest

from petlib.bn import Bn

from sscred.commitment import *
from sscred.blind_signature import *
from sscred.blind_pedersen import *
from sscred.acl import *
from sscred.config import DEFAULT_GROUP_ID
from sscred.pack import packb, unpackb, add_msgpack_support


def test_pedersen_commit_verification():
    values = [Bn(2651), Bn(1), Bn(98)]
    pparam = PedersenParameters(hs_size=len(values))
    assert pparam.verify_parameters()
    pcommit, prand = pparam.commit(values)
    assert pcommit.verify(pparam, prand, values)


def test_pedersen_commit_proof():
    values = [Bn(2651), Bn(1), Bn(98)]
    pparam = PedersenParameters(hs_size=len(values))
    pcommit, prand = pparam.commit(values)
    proof = pcommit.prove_knowledge(pparam, prand, values)
    assert pcommit.verify_proof(pparam, proof)


def test_pedersen_commit_invalid_proof():
    values = [Bn(2651), Bn(1), Bn(98)]
    pparam = PedersenParameters(hs_size=len(values))
    pcommit, prand = pparam.commit(values)
    prand = PedersenRandom(pparam.q.random())
    assert not pcommit.verify(pparam, prand, values)
    proof = pcommit.prove_knowledge(pparam, prand, values)
    assert not pcommit.verify_proof(pparam, proof)


def auto_sign(signer: AbeSigner, user: AbeUser, message: Union[bytes, str]) -> AbeSignature:
    m1, p1 = signer.commit()
    m2, p2 = user.compute_blind_challenge(m1, message)
    m3 = signer.respond(m2, p1)
    sig = user.compute_signature(m3, p2)
    return sig


def test_abe_signature_verification():
    priv, pk = AbeParam().generate_new_key_pair()

    # To ensure that signer's previous signature's state does not corrupt future
    # queries if used in an iterative way.
    signer = AbeSigner(priv, pk)
    user = AbeUser(pk)
    for i in range(1, 10):
        message = f"Hello world{i}"
        sig = auto_sign(signer, user, message)
        assert pk.verify_signature(sig)
        assert message.encode('utf8') == sig.message


def test_abe_signature_concurrent_fail_if_acl_is_enabled():
    priv, pk = AbeParam().generate_new_key_pair()
    signer = AbeSigner(priv, pk, disable_acl=False)

    signer.commit()
    with pytest.raises(AbeSignerStateInvalid):
        signer.commit()


def test_abe_signature_concurrent_fail_if_acl_is_disabled():
    priv, pk = AbeParam().generate_new_key_pair()
    signer = AbeSigner(priv, pk, disable_acl=True)

    signer.commit()
    signer.commit()


def test_abe_signature_verification_corrupted():
    priv, pk = AbeParam().generate_new_key_pair()
    priv2, pk2 = AbeParam().generate_new_key_pair()
    message = "Hello world"

    # verifying against wrong public key
    signer = AbeSigner(priv, pk)
    user = AbeUser(pk)
    sig = auto_sign(signer, user, message)
    assert not pk2.verify_signature(sig)

    # wrong secret key
    signer = AbeSigner(priv2, pk)
    user = AbeUser(pk)
    sig = auto_sign(signer, user, message)
    assert not pk.verify_signature(sig)

    # wrong public key for the user
    signer = AbeSigner(priv, pk)
    user = AbeUser(pk2)
    sig = auto_sign(signer, user, message)
    assert not pk.verify_signature(sig)
    assert not pk2.verify_signature(sig)


def test_bl_pedersen_valid():
    values = [Bn(123), Bn(456), 'hello', b"world"]
    param = BlindedPedersenParam(hs_size=len(values))

    # reveal nothing
    bcommit, bpriv = param.blind_commit(values)
    bproof = bcommit.prove_values(bpriv)
    assert bcommit.verify_proof(param, bproof)

    # revealing some values
    bproof2 = bcommit.prove_values(bpriv, reveal_mask=[True, False, True, True])
    assert bcommit.verify_proof(param, bproof2)
    assert bproof2.revealed_values == [Bn(123), None, 'hello', b"world"]


def test_bl_pedersen_with_acl_params_valid():
    group = EcGroup(DEFAULT_GROUP_ID)
    g1 = group.hash_to_point(b"test_z")
    g2 = group.hash_to_point(b"test_h2")

    values = [Bn(123), Bn(456), 'hello', b"world"]
    param = BlindedPedersenParam(hs_size=len(values), Z=g1, H_2=g2 )
    bcommit, bpriv = param.blind_commit(values)

    bproof = bcommit.prove_values(bpriv, reveal_mask=[True, False, True, True])
    assert bcommit.verify_proof(param, bproof)
    assert bproof.revealed_values == [Bn(123), None, 'hello', b"world"]


def test_bl_pedersen_invalid():
    group = EcGroup(DEFAULT_GROUP_ID)
    g1 = group.hash_to_point(b"test_z")
    g2 = group.hash_to_point(b"test_h2")

    values = [Bn(123), Bn(456), 'hello', b"world"]
    param = BlindedPedersenParam(hs_size=len(values), Z=g1, H_2=g2)
    bcommit, bpriv = param.blind_commit(values)

    # corrupt rand
    bpriv2 = unpackb(packb(bpriv))
    bpriv2.rand.value = param.q.random()
    bproof = bcommit.prove_values(bpriv2, reveal_mask=[True, False, True, True])
    assert not bcommit.verify_proof(param, bproof)

    # corrupt rand2
    bpriv3 = unpackb(packb(bpriv))
    bpriv3.rand_2.value = param.q.random()
    bproof = bcommit.prove_values(bpriv3, reveal_mask=[True, False, True, True])
    assert not bcommit.verify_proof(param, bproof)

    # corrupt blindness
    bpriv4 = unpackb(packb(bpriv))
    bpriv4.blindness_rand.value = param.q.random()
    bproof = bcommit.prove_values(bpriv4, reveal_mask=[True, False, True, True])
    assert not bcommit.verify_proof(param, bproof)

    # corrupt revealed value
    # This leads to different zksk.statement constants between prover and
    # verifier which result in StatementMismatch exception. verify_proof checks
    # for this statement and convert it to False
    bpriv5 = unpackb(packb(bpriv))
    bpriv5.values[0].value = param.q.random()
    bproof = bcommit.prove_values(bpriv5, reveal_mask=[True, False, True, True])
    assert not bcommit.verify_proof(param, bproof)

    # corrupt hidden value
    bpriv6 = unpackb(packb(bpriv))
    bpriv6.values[1].value = param.q.random()
    bproof = bcommit.prove_values(bpriv6, reveal_mask=[True, False, True, True])
    assert not bcommit.verify_proof(param, bproof)


def auto_cred(user: ACLUser, issuer: ACLIssuer, attrs: List[Attribute]):
    # Interactive signing
    message = "This isn't a test message."
    m0 = user.prove_attr_knowledge(attrs)
    m1, p1 = issuer.commit(m0)
    m2, p2 = user.compute_blind_challenge(m1, message)
    m3 = issuer.respond(m2, p1)
    return user.compute_credential(m3, p2)


def test_acl_valid():
    # generating keys and wrappers
    issuer_priv, issuer_pk = ACLParam().generate_new_key_pair()
    issuer = ACLIssuer(issuer_priv, issuer_pk)
    user = ACLUser(issuer_pk)

    attrs = [Bn(13), "Hello", "WoRlD", "Hidden"]
    cred_private = auto_cred(user, issuer, attrs)

    # show credential
    cred = cred_private.show_credential([True, True, True, False])
    assert cred.verify_credential(issuer_pk)
    assert cred.message() == b"This isn't a test message."
    assert cred.attributes() == [13, 'Hello', 'WoRlD', None]

    with pytest.raises(Exception):
        cred_private.show_credential([True, False, True, False])


def test_acl_invalid():
    # generating keys and wrappers
    issuer_priv, issuer_pk = ACLParam().generate_new_key_pair()
    issuer_priv2, issuer_pk2 = ACLParam().generate_new_key_pair()
    attrs = [Bn(13), "Hello", "WoRlD", "Hidden"]

    # verifying against wrong public key
    issuer = ACLIssuer(issuer_priv, issuer_pk)
    user = ACLUser(issuer_pk)
    cred_private = auto_cred(user, issuer, attrs)
    cred = cred_private.show_credential([True, True, True, False])
    assert not cred.verify_credential(issuer_pk2)

    # wrong secret key
    issuer = ACLIssuer(issuer_priv2, issuer_pk)
    user = ACLUser(issuer_pk)
    cred_private = auto_cred(user, issuer, attrs)
    cred = cred_private.show_credential([True, True, True, False])
    assert not cred.verify_credential(issuer_pk)

    # wrong public key for the issuer
    issuer = ACLIssuer(issuer_priv, issuer_pk2)
    user = ACLUser(issuer_pk)
    cred_private = auto_cred(user, issuer, attrs)
    cred = cred_private.show_credential([True, True, True, False])
    assert not cred.verify_credential(issuer_pk)
    assert not cred.verify_credential(issuer_pk2)

    # wrong public key for the user
    issuer = ACLIssuer(issuer_priv, issuer_pk)
    user = ACLUser(issuer_pk2)
    cred_private = auto_cred(user, issuer, attrs)
    cred = cred_private.show_credential([True, True, True, False])
    assert not cred.verify_credential(issuer_pk)
    assert not cred.verify_credential(issuer_pk2)

    # reveal with partial mask
    issuer = ACLIssuer(issuer_priv, issuer_pk)
    user = ACLUser(issuer_pk)
    cred_private = auto_cred(user, issuer, attrs)
    with pytest.raises(Exception):
        cred = cred_private.show_credential([True, True])


def test_pack_cls():
    @attr.s
    class CLS(object):
        a = attr.ib()
        b = attr.ib()
    add_msgpack_support(CLS, 10)

    obj = CLS(a=[3, 'asd', Bn(123)], b=EcGroup().generator())
    obj2 = CLS(a=False, b=11 * EcGroup().generator())
    d = obj.to_bytes()
    objp = CLS.from_bytes(d)
    assert obj == objp

    x = [obj, obj2]
    bt = packb(x)
    xp = unpackb(bt)
    assert x == xp


def test_pack_blind_sig():
    priv, pk = AbeParam().generate_new_key_pair()
    signer = AbeSigner(priv, pk)
    user = AbeUser(pk)
    message = "Hello world"
    com, signer_params = signer.commit()
    challenge, user_params = user.compute_blind_challenge(com, message)
    resp = signer.respond(challenge, signer_params)
    sig = user.compute_signature(resp, user_params)

    m1 = packb(com)
    m2 = packb(challenge)
    m3 = packb(resp)
    m4 = packb(sig)
    m5 = packb(priv)
    m6 = packb(pk)

    assert unpackb(m1) == com
    assert unpackb(m2) == challenge
    assert unpackb(m3) == resp
    assert unpackb(m4) == sig
    assert unpackb(m5) == priv
    assert unpackb(m6) == pk


def test_pack_acl():
    issuer_priv, issuer_pk = ACLParam().generate_new_key_pair()
    issuer = ACLIssuer(issuer_priv, issuer_pk)
    user = ACLUser(issuer_pk)
    attrs = [Bn(13), "Hello", "WoRlD", "Hidden"]
    message = "This isn't a test message."

    m0 = user.prove_attr_knowledge(attrs)
    m1, p1 = issuer.commit(m0)
    m2, p2 = user.compute_blind_challenge(m1, message)
    m3 = issuer.respond(m2, p1)
    cred_private = user.compute_credential(m3, p2)
    cred = cred_private.show_credential([True, True, True, False])

    m0p = packb(m0)
    m1p = packb(m1)
    m2p = packb(m2)
    m3p = packb(m3)
    cred_privatep = packb(cred_private)
    credp = packb(cred)


    # the original message is representing m0.nizk_proof.response as list while
    # the unpacked version represent it as a tuple
    # assert unpackb(m0p) == m0

    assert unpackb(m1p) == m1
    assert unpackb(m2p) == m2
    assert unpackb(m3p) == m3
    assert unpackb(cred_privatep) == cred_private
    assert unpackb(credp) == cred

    cred = unpackb(credp)
    assert cred.verify_credential(issuer_pk)
    assert cred.message() == b"This isn't a test message."
    assert cred.attributes() == [13, 'Hello', 'WoRlD', None]


def main():
    pass


if __name__ == '__main__':
    main()



"""
Abe's blind signature scheme
Ref: # Abe, M. A Secure Three-move Blind Signature Scheme for Polynomially
Many Signatures.

An implementation of Abe's blind signature. An user 'AbeUser' asks a signer
'AbeSigner' to sign a message 'message' without revealing its content.

The parameters' names follow the Abe's paper notation.

Example:
    >>> # generating keys and wrappers
    >>> priv, pk = AbeParam().generate_new_key_pair()
    >>> signer = AbeSigner(priv, pk)
    >>> user = AbeUser(pk)
    >>> message = b"Hello world"
    >>> # Interactive signing
    >>> com, com_intern = signer.commit()
    >>> challenge, challenge_intern = user.compute_blind_challenge(com, message)
    >>> resp = signer.respond(challenge, com_intern)
    >>> sig = user.compute_signature(resp, challenge_intern)
    >>> # Verifying the signature
    >>> assert pk.verify_signature(sig)
    >>> print(sig.message)
    b'Hello world'
"""

from __future__ import annotations

from enum import IntEnum
from hashlib import sha256
from threading import Lock
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
    Union,
)

import attr
from petlib.bn import Bn
from petlib.ec import (
    EcPt,
    EcGroup,
)

from . import config


class AbeSignerStateInvalid(Exception):
    """The Abe signer is in an invalid state."""


class AbePublicKeyInvalid(Exception):
    """The public key for Abe's blind signature scheme is invalid."""


class AbeSignerState(IntEnum):
    """State of the Abe signer."""
    COMMITTED = 0
    READY_TO_COMMIT = 1


class AbeParam:
    """Parameters for Abe's blind signature scheme.

    :param group: elliptic curve to use for the blind signature
    """

    __slots__ = ("group", "q", "g", "h")

    def __init__(self, group: Optional[EcGroup] = None) -> None:
        self.group: EcGroup = EcGroup(config.DEFAULT_GROUP_ID) if group is None else group
        self.q: Bn = self.group.order()
        self.g: EcPt = self.group.hash_to_point(b"sig_g")
        self.h: EcPt = self.group.hash_to_point(b"sig_h")


    def generate_new_key_pair(self) -> Tuple[AbePrivateKey, AbePublicKey]:
        """Generate a new key pair for Abe's blind signature.

        :return: a tuple containing a private key and its corresponding public key
        """
        sk: Bn = self.q.random()
        private = AbePrivateKey(sk)
        public = private.public_key(self)
        return private, public


@attr.s(slots=True)
class AbeSignature:
    message = attr.ib() # type: Union[bytes, str]
    zeta = attr.ib()    # type: EcPt
    zeta1 = attr.ib()   # type: EcPt
    rho = attr.ib()     # type: Bn
    w = attr.ib()       # type: Bn
    delta1 = attr.ib()  # type: Bn
    delta2 = attr.ib()  # type: Bn
    sigma = attr.ib()   # type: Bn
    micro = attr.ib()   # type: Bn


@attr.s(slots=True)
class SignerCommitMessage:
    rnd = attr.ib() # type: Bn
    a = attr.ib()   # type: EcPt
    b1 = attr.ib()  # type: EcPt
    b2 = attr.ib()  # type: EcPt


@attr.s(slots=True)
class SignerCommitmentInternalState:
    u = attr.ib()  # type: Bn
    s1 = attr.ib() # type: Bn
    s2 = attr.ib() # type: Bn
    d = attr.ib()  # type: Bn

    @classmethod
    def new(cls, q: Bn) -> SignerCommitmentInternalState:
        """Generate a new set of random internal parameters."""
        return cls(*(q.random() for _ in range(4)))


@attr.s(slots=True)
class BlindedChallengeMessage:
    e = attr.ib() # type: Bn


@attr.s(slots=True)
class UserBlindedChallengeInternalState:
    blinder = attr.ib() # type: Bn
    tau = attr.ib()     # type: Bn
    t = attr.ib()       # type: List[Bn]
    zeta = attr.ib()    # type: EcPt
    zeta1 = attr.ib()   # type: EcPt
    message = attr.ib() # type: bytes


@attr.s(slots=True)
class SignerResponseMessage:
    r = attr.ib()  # type: Bn
    c = attr.ib()  # type: Bn
    s1 = attr.ib() # type: Bn
    s2 = attr.ib() # type: Bn
    d = attr.ib()  # type: Bn


@attr.s(slots=True)
class AbePrivateKey:
    sk = attr.ib() # type: Bn

    def public_key(self, params: Optional[AbeParam] = None) -> AbePublicKey:
        """Create a public key corresponding to this private key."""
        if params is None:
            params = AbeParam()
        return AbePublicKey(params, self)


class AbePublicKey:
    """Public key for Abe's blind signature scheme.

    :param param: parameters
    :param priv: signer's private key
    """

    __slots__ = ("param", "pk", "z")


    def __init__(self, param: AbeParam, priv: AbePrivateKey) -> None:
        self.param: AbeParam = param
        self.pk: EcPt = priv.sk * param.g
        self.z: EcPt = self._compute_z_param()


    def _compute_z_param(self) -> EcPt:
        return self.param.group.hash_to_point(
            self.param.g.export() + self.param.h.export() + self.pk.export()
        )


    def verify_parameters(self, verify_bases: bool = False) -> bool:
        """Verify that the public key is valid.

        :param verify_bases: verifies public parameter's generators G and H
        :return: True if the public key is valid, False otherwise
        """
        if verify_bases:
            if self.param.g != self.z.group.hash_to_point(b"sig_g"):
                return False
            if self.param.h != self.z.group.hash_to_point(b"sig_h"):
                return False
        return self.z == self._compute_z_param()


    def verify_signature(self, sig: AbeSignature) -> bool:
        """Verify that the signature is valid

        :param sig: the signature to verify
        :return: True if the signature is valid, False otherwise
        """

        param: AbeParam = self.param

        if not all((
            param.group.check_point(sig.zeta),
            param.group.check_point(sig.zeta1),
            (0 <= sig.rho < param.q),
            (0 <= sig.w < param.q),
            (0 <= sig.delta1 < param.q),
            (0 <= sig.delta2 < param.q),
            (0 <= sig.sigma < param.q),
            (0 <= sig.micro < param.q),
            isinstance(sig.message, bytes),
            not sig.zeta.is_infinite(),
        )):
            return False

        zeta2: EcPt = sig.zeta - sig.zeta1
        h = sha256(
            b'||'.join((
                sig.zeta.export(),
                sig.zeta1.export(),
                (sig.rho * param.g + sig.w * self.pk).export(),
                (sig.delta1 * param.g + sig.sigma * sig.zeta1).export(),
                (sig.delta2 * param.h + sig.sigma * zeta2).export(),
                (sig.micro * self.z + sig.sigma * sig.zeta).export(),
                sig.message,
            ))
        ).digest()

        lhs: Bn = (sig.w + sig.sigma) % param.q
        rhs: Bn = Bn.from_binary(h) % param.q

        return lhs == rhs


class AbeSigner:
    """Signer for Abe's blind signature scheme

    **Warning:** When used for Anonymous credentials light (ACL), commitments can not be issued in
    parallel. Instead, the signer and the user have to issue commitment and respond back a blinded
    challenge sequentially. Therefore the signer can not issue a new commitment until the user
    responded to the previous commitment.

    By default the signer will ensure it can not issue a commitment until the previous one was
    responded correctly, and these checks are thread safe. Optionally, you can disable the thread
    safety meachanism if you prefer to implement it at a higher level, or completely disable the
    safety checks if you do not intent to use the signer for ACL.

    :param private: signer's private key
    :param public: signer's public key
    :param disable_acl: disable check to ensure ACL validity
    :param thread_safe: ensure the signer is thread safe (parameter ignored when ACL usage is
        disabled)
    """

    __slots__ = ("param", "public", "private", "enable_acl", "state", "lock")

    def __init__(
            self,
            private: AbePrivateKey,
            public: AbePublicKey,
            disable_acl: bool = False,
            thread_safe: bool = True
        ) -> None:
        self.param: AbeParam = public.param
        self.public: AbePublicKey = public
        self.private: AbePrivateKey = private
        self.enable_acl: bool = not disable_acl
        self.state: AbeSignerState = AbeSignerState.READY_TO_COMMIT
        self.lock: Lock = Lock() if self.enable_acl and thread_safe else None


    def commit(self) -> Tuple[SignerCommitMessage, SignerCommitmentInternalState]:
        """Initiate the signing protocol.

        :raises AbeSignerStateInvalid: The signer attempted to issue a new commitment before
            receiving a blinded challenge for the previous one.
        :return: a tuple containing a commitment message and the commitment's internal parameters
            that will be necessary to process the user's blinded challenge to this commitment
        """
        group = self.param.group
        return self._commit(lambda rnd, group=group: group.hash_to_point(b"z1_" + rnd.binary()))


    def _commit(
            self,
            compute_z1: Callable[[Bn], EcPt]
        ) -> Tuple[SignerCommitMessage, SignerCommitmentInternalState]:

        if self.enable_acl:
            if self.lock is not None:
                self.lock.acquire()
            if self.state != AbeSignerState.READY_TO_COMMIT:
                raise AbeSignerStateInvalid(
                    "Can not make a new commitment until a blinded challenge is received for the "
                    "previous one."
                )

        rnd: Bn = self.param.q.random()
        z1: EcPt = compute_z1(rnd)
        z2: EcPt = self.public.z - z1

        rnd_params = SignerCommitmentInternalState.new(self.param.q)

        a: EcPt = rnd_params.u * self.param.g
        b1: EcPt = rnd_params.s1 * self.param.g + rnd_params.d * z1
        b2: EcPt = rnd_params.s2 * self.param.h + rnd_params.d * z2

        if self.enable_acl:
            self.state = AbeSignerState.COMMITTED
            if self.lock is not None:
                self.lock.release()

        return (
            SignerCommitMessage(rnd, a, b1, b2),
            rnd_params,
        )


    def respond(
            self,
            challenge: BlindedChallengeMessage,
            commit_state: SignerCommitmentInternalState
        ) -> SignerResponseMessage:
        """Compute the response for the user's challenge.

        :param challenge: the user's blinded challenge
        :param internal_params: internal parameters of the commitment to which the user responded
            with this blinded challenge
        :raises AbeSignerStateInvalid: The signer attempted to respond to a blinded challenge
            before issuing a commitment.
        :return: response to the blinded message
        """
        if self.enable_acl:
            if self.lock is not None:
                self.lock.acquire()
            if self.state != AbeSignerState.COMMITTED:
                raise AbeSignerStateInvalid(
                    "Attempt to respond to a commitment which has not been issued."
                )

        c: Bn = (challenge.e - commit_state.d) % self.param.q
        r: Bn = (commit_state.u - c * self.private.sk) % self.param.q

        if self.enable_acl:
            self.state = AbeSignerState.READY_TO_COMMIT
            if self.lock is not None:
                self.lock.release()

        return SignerResponseMessage(
            r,
            c,
            commit_state.s1,
            commit_state.s2,
            commit_state.d
        )


class AbeUser:
    """User for Abe's blind signature scheme.

    :param public: signer's public key
    :param verify_pk: verify the signer's public key
    :raises AbePublicKeyInvalid: The public key is invalid
    """

    __slots__ = ("param", "public")

    def __init__(self, public: AbePublicKey, verify_pk: bool = False):
        self.param: AbeParam = public.param
        self.public: AbePublicKey = public
        if verify_pk and not self.public.verify_parameters():
            raise AbePublicKeyInvalid()


    def compute_blind_challenge(
            self,
            commit_message: SignerCommitMessage,
            message: Union[bytes, str]
        ) -> Tuple[BlindedChallengeMessage, UserBlindedChallengeInternalState]:
        """Compute a blind challenge with a commitment from teh signer and a message to sign.

        :param commit_message: commitment from the signer
        :param message: a message to sign
        :raises ValueError: The commitment or the message to sign is invalid.
        :return: a tuple containing a blind challenge to respond back to the signer and the blind
            challenge's internal parameters that will be necessary to process the signer's response
            to this blind challenge
        """
        group = self.param.group
        return self._compute_blind_challenge(
            commit_message,
            message,
            lambda rnd, group=group: group.hash_to_point(b"z1_" + rnd.binary())
        )


    def _compute_blind_challenge(
            self,
            commit_message: SignerCommitMessage,
            message: Union[bytes, str],
            compute_z1: Callable[[Bn], EcPt]
        ) -> Tuple[BlindedChallengeMessage, UserBlindedChallengeInternalState]:
        if isinstance(message, str):
            message = message.encode('utf8')
        if not isinstance(message, bytes):
            raise ValueError("Invalid message, 'message' is not bytes")

        rnd: Bn = commit_message.rnd
        z1: EcPt = compute_z1(rnd)

        a: EcPt
        b1: EcPt
        b2: EcPt
        a, b1, b2 = (
            commit_message.a,
            commit_message.b1,
            commit_message.b2,
        )

        if not self.param.group.check_point(a):
            raise ValueError("Invalid point A")
        if not self.param.group.check_point(b1):
            raise ValueError("Invalid point B1")
        if not self.param.group.check_point(b2):
            raise ValueError("Invalid point B2")

        blinder: Bn = self.param.q.random()
        tau: Bn = self.param.q.random()
        t: List[Bn] = [self.param.q.random() for _ in range(5)]
        zeta: EcPt = blinder * self.public.z
        zeta1: EcPt = blinder * z1

        eta: EcPt = tau * self.public.z
        zeta2: EcPt = zeta - zeta1

        alpha: EcPt = a + t[0] * self.param.g + t[1] * self.public.pk
        beta1: EcPt = blinder * b1 + t[2] * self.param.g + t[3] * zeta1
        beta2: EcPt = blinder * b2 + t[4] * self.param.h + t[3] * zeta2

        h: bytes = sha256(b'||'.join(
            (
                zeta.export(),
                zeta1.export(),
                alpha.export(),
                beta1.export(),
                beta2.export(),
                eta.export(),
                message
            )
        )).digest()

        epsilon: Bn = Bn.from_binary(h) % self.param.q

        e: Bn = (epsilon - t[1] - t[3]) % self.param.q
        return (
            BlindedChallengeMessage(e),
            UserBlindedChallengeInternalState(blinder, tau, t, zeta, zeta1, message)
        )


    def compute_signature(
            self,
            response: SignerResponseMessage,
            challenge_state: UserBlindedChallengeInternalState
        ) -> AbeSignature:
        """Finish the blind signing's sigma protocol and compute a signature

        :param response: signer's response to the blind challenge
        :param challenge_private: internal parameters of the blind challenge to which the signer
            responded.
        :return: a signature which can be verified agains the signer's public key
        """
        rho: Bn = (response.r + challenge_state.t[0]) % self.param.q
        w: Bn = (response.c + challenge_state.t[1]) % self.param.q
        delta1: Bn = (challenge_state.blinder * response.s1 + challenge_state.t[2]) % self.param.q
        delta2: Bn = (challenge_state.blinder * response.s2 + challenge_state.t[4]) % self.param.q
        sigma: Bn = (response.d + challenge_state.t[3]) % self.param.q
        micro: Bn = (challenge_state.tau - sigma * challenge_state.blinder) % self.param.q

        return AbeSignature(
            challenge_state.message,
            challenge_state.zeta,
            challenge_state.zeta1,
            rho,
            w,
            delta1,
            delta2,
            sigma,
            micro,
        )


def main():
    import doctest
    doctest.testmod(verbose=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import os
import re

from setuptools import setup

PACKAGE_NAME = "sscred"
INSTALL_REQUIRES = [
    "petlib_fork",
    # "petlib @ git+https://github.com/spring-epfl/petlib.git",
    "attrs",
    # "zksk"
    "zksk_fork"
]
SETUP_REQUIRES = ["pytest-runner"]
TEST_REQUIRES = ["pytest"]


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.md")) as f:
    long_description = f.read()


with open(os.path.join(here, PACKAGE_NAME, "__init__.py")) as f:
    matches = re.findall(r"(__.+__) = \"(.*)\"", f.read())
    for var_name, var_value in matches:
        globals()[var_name] = var_value

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    globals()['long_description'] = f.read()


setup(
    name="sscred_fork",
    packages=[PACKAGE_NAME],
    install_requires=INSTALL_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    tests_require=TEST_REQUIRES,
    extras_require={"test": TEST_REQUIRES},
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Security :: Cryptography",
        "License :: OSI Approved :: BSD License",
    ],
)



"""
An implementation of blinded Pedersen commitment and utility support for using
it. Features: commit, blind_commit, partial open, nizk proof, verify

Examples:
    >>> values = [Bn(123), Bn(456), 'hello', b"world"]
    >>> param = BlindedPedersenParam(hs_size=len(values))

    >>> # reveal nothing
    >>> bcommit, bpriv = param.blind_commit(values)
    >>> bproof = bcommit.prove_values(bpriv)
    >>> assert bcommit.verify_proof(param, bproof)

    >>> # revealing some values
    >>> bproof = bcommit.prove_values(bpriv, reveal_mask=[True, False, True, True])
    >>> assert bcommit.verify_proof(param, bproof)
    >>> print(bproof.revealed_values)
    [123, None, 'hello', b'world']
"""

from hashlib import sha256
from typing import List, Optional, Union

import attr

from petlib.bn import Bn
from petlib.ec import EcGroup, EcPt

from zksk import Secret, DLRep
from zksk.base import NIZK
from zksk.exceptions import StatementMismatch

from .commitment import PedersenParameters


class BlindedPedersenParam(PedersenParameters):
    """ Common parameters for blinded Pedersen.
    This class provides both vanilla and blinded Pedersen commits. The vanilla
    commit is compatible with PedersenCommitment.
    """

    def __init__(self, group: Optional[EcGroup] = None, hs_size=3, Z=None, H_2=None):
        """ Return the common parameters for the specified curve.

        Z and H_2 are intended for compatibilities between SSCred submodules.
        The ACL submodule uses these two arguments to customize the commitment.
        For standalone use of the commitment, leave Z and H_2 empty.

        Args:
            group (EcGroup): determines the elliptic curve used in the scheme.
            hs_size (int): maximum number of attributes allowed in the commitment.
            Z (EcPt): blindness generator(z^y, C^y) (Optional)
            H_2 (EcPt): Additional randomizer. Allows ACl compatibility. (Optional)
        """
        super().__init__(group, hs_size)
        self.Z = Z if Z is not None else self.group.hash_to_point(b"bl_z")
        self.H_2 = H_2 if H_2 is not None else self.group.infinite()

    def process_raw_value(self, raw):
        """"process raw values.

        Args:
            raw_values (int, Bn mod q/string/bytes): committed value.
        Raises:
            Error: bad value encoding. values can only be Bn, bytes, string,
                or None.
        Returns:
            (Bn mod q): processed value
        """
        val = None
        if raw is None:
            return None
        elif isinstance(raw, Bn):
            return raw
        elif isinstance(raw, bytes):
            val = sha256(raw).digest()
        elif isinstance(raw, str):
            val = sha256(raw.encode('utf8')).digest()
        elif isinstance(raw, int):
            return Bn(raw)
        else:
            raise Exception("Bad commitment value encoding."
                "values can only be int, Bn, bytes, string, or None.")
        return Bn.from_binary(val) % self.q


    def process_raw_values(self, raw_values):
        """process raw values.

        Args:
            raw_values (int, Bn mod q/string/bytes[]): committed value.
        Raises:
            Error: bad value encoding. values can only be Bn, bytes, string,
                or None.

        Returns:
            (Bn mod q []): processed value
        """
        return [self.process_raw_value(raw) for raw in raw_values]


    def commit(self, raw_values):
        """"A non-blinded Pedersen commitment to values. Encodes raw values before
        passing them to the PedersenCommitment.

        IMPORTANT: THIS FUNCTION IS NOT BLINDED.

        Args:
            raw_values (Bn mod q/string/bytes []): committed values.
        Raises:
            Error: bad value encoding. values can only be Bn, bytes, or string.
            Error: too many values

        Returns:
            (PedersenCommitment): commitment
            (PedersenRand): commitment's randomness
        """

        values = [self.process_raw_value(raw) for raw in raw_values]
        return super(BlindedPedersenParam, self).commit(values)


    def blind_commit(self, raw_values, blindness_rand=None, h_rand=None):
        """"a blind commitment to values.

        Arguments blindness_rand and h_rand enable ACL compatability.

        Args:
            raw_values (Bn mod q/string/bytes []): committed values.
            blindness_rand (Bn mod q): forces the blinding's randomness. (optional)
            h_rand (Bn): fix the randomness for H. (optional)
        Raises:
            Error: bad value encoding. values can only be Bn, bytes, or string.
            Error: blindness generator is unknown

        Returns:
            BlPedersenCommitment: commitment
            BlPedersenPrivate: commitment's values and randomness
        """

        values = [self.process_raw_value(raw) for raw in raw_values]

        if len(values) > len(self.HS):
            self.expand_params_len(len(values))

        rand = h_rand if h_rand is not None else self.q.random()

        rand_2 = self.q.random()
        if blindness_rand is None:
            blindness_rand = self.q.random()

        C = (rand * self.H) + (rand_2 * self.H_2)
        C += self.group.wsum(values, self.HS[: len(values)])

        bl_commit = blindness_rand * C
        bl_z = blindness_rand * self.Z

        bcommit = BlPedersenCommitment(bl_z=bl_z, bl_commit=bl_commit)
        bpriv = BlPedersenPrivate(
            param=self,
            raw_values=list(raw_values),
            values=[Secret(val, name=f"val_{i}") for (i, val) in enumerate(values)],
            rand=Secret(rand, name="rand"),
            rand_2=Secret(rand_2, name="rand_2"),
            blindness_rand=Secret(blindness_rand, name="blindness_rand"),
        )

        return (bcommit, bpriv)


@attr.s
class BlPedersenPrivate():
    """A blinded Pedersen commitment's secret values.

    Attributes:
        param (BlindedPedersenParam): commitment parameters
        raw_values (Union[Bn,string,bytes]): raw committed values.
        values (Secret[]): committed values.
        rand (Secret): BlPedersen's commit randomness.
        rand_2 (Secret): extra commit randomness for ACL.
        blindness_rand (Secret): blindness randomness
    """
    param = attr.ib()          # type: BlindedPedersenParam
    raw_values = attr.ib()     # type: Union[Bn,str,bytes]
    values = attr.ib()         # type: List[Secret]
    rand = attr.ib()           # type: Secret
    rand_2 = attr.ib()         # type: Secret
    blindness_rand = attr.ib() # type: Secret


@attr.s
class BlPedersenProof():
    """A NIZK proof of knowing the opening of a BlindedPedersenCommitment.

    Attributes:
        bl_h (EcPt): blindness for H
        bl_h2 (EcPt): blindness for H_2
        bl_hi (EcPt[]): blindness for HS[i]
        revealed_values (Bn []): a list of all values. If a value is not
            revealed then rv[i]=None, otherwise rv[i]=open(value[i])
        proof (zksk.NIZK): zksk's nizk proof

    Reminder:
        The verifier already has bl_z and bl_commit from the commit phase
    """
    bl_h = attr.ib()              # type: EcPt
    bl_h2 = attr.ib()             # type: EcPt
    bl_hi = attr.ib()             # type: EcPt
    revealed_values = attr.ib()   # type: List[Bn]
    nizk_proof = attr.ib()        # type: NIZK


class BlPedersenCommitment:
    """The public commitment in a blinded Pedersen commitment"""

    def __init__(self, bl_z, bl_commit):
        """Create a blinded Pedersen commitment.

        Attributes:
            bl_commit (EcPt): randomized commitment = C ^ blindness_rand
            bl_z (EcPt): randomized Z base = Z ^ blindness_rand
        """
        self.bl_commit = bl_commit
        self.bl_z = bl_z

    def prove_values(self, bpriv, reveal_mask=None):
        """A nizk proof of opening with the revealed values.
        The proof contains the raw value of all revealed values. This protocol
        does not reveal any information about non-revealed values.

        Error:
            Either reveal_mask should be None or have the same size of private values.
        Args:
            bpriv (BlPedersenPrivate): values and randomness.
            reveal_mask (boolean []): reveals value[i] iff reveal_mask[i]
                Optional: mask=None -> reveal nothing

        Returns: (BlPedersenProof): a nizk proof for commitment (self)
        """
        param = bpriv.param
        if reveal_mask is None:
            reveal_mask = len(bpriv.values) * [False]
        if len(reveal_mask) != len(bpriv.values):
            raise Exception('The size of reveal mask does not match the number of attributes')

        # randomness
        expr_bl_z = bpriv.blindness_rand * param.Z
        expr_bl_h = bpriv.blindness_rand * param.H
        expr_bl_h2 = bpriv.blindness_rand * param.H_2
        bl_h, bl_h2 = expr_bl_h.eval(), expr_bl_h2.eval()

        stmt = (
            DLRep(self.bl_z, expr_bl_z) &
            DLRep(bl_h, expr_bl_h) &
            DLRep(bl_h2, expr_bl_h2)
        )

        # attributes
        bl_hi = list()
        for i in range(len(bpriv.values)):
            expr = bpriv.blindness_rand * param.HS[i]
            bl_hi.append(expr.eval())
            stmt = stmt & DLRep(bl_hi[i], expr)

        # proof
        revealed_values = list()
        revealed_acc = param.group.infinite()
        expr_bl_commit = bpriv.rand * bl_h + bpriv.rand_2 * bl_h2
        for i in range(len(bpriv.values)):
            if reveal_mask[i]:
                revealed_values.append(bpriv.raw_values[i])
                revealed_acc += bpriv.values[i].value * bl_hi[i]
            else:
                revealed_values.append(None)
                expr_bl_commit += bpriv.values[i] * bl_hi[i]
        stmt = stmt & DLRep(self.bl_commit - revealed_acc, expr_bl_commit)
        bproof = BlPedersenProof(
            bl_h=bl_h,
            bl_h2=bl_h2,
            bl_hi=bl_hi,
            revealed_values=revealed_values,
            nizk_proof=stmt.prove()
        )
        return bproof

    def verify_proof(self, bc_param, bproof):
        """Verify a PedersenProof for this commitment.
        Important: besides verifying the proof, you should verify the parameters.

        Args:
            bc_param (BlPedersenParam): commitment parameters
            bproof (BlPedersenProof): a nizk proof for self
        Returns:
            boolean: pass/fail
        """

        if bc_param.num_params() < len(bproof.bl_hi):
            bc_param.expand_params_len(len(bproof.bl_hi))
        bl_rand_sec = Secret(name="blindness_rand")

        # randomness
        expr_bl_z = bl_rand_sec * bc_param.Z
        expr_bl_h = bl_rand_sec * bc_param.H
        expr_bl_h2 = bl_rand_sec * bc_param.H_2
        stmt = (
            DLRep(self.bl_z, expr_bl_z) &
            DLRep(bproof.bl_h, expr_bl_h) &
            DLRep(bproof.bl_h2, expr_bl_h2)
        )

        # attributes
        for i in range(len(bproof.bl_hi)):
            expr = bl_rand_sec *bc_param.HS[i]
            stmt = stmt & DLRep(bproof.bl_hi[i], expr)

        # proof
        revealed_acc = bc_param.group.infinite()
        expr_bl_commit = Secret(name="rand") * bproof.bl_h + Secret(name="rand_2") * bproof.bl_h2
        for i in range(len(bproof.bl_hi)):
            if bproof.revealed_values[i] is None:
                expr_bl_commit += Secret(name=f"val_{i}") * bproof.bl_hi[i]
            else:
                val = bc_param.process_raw_value(bproof.revealed_values[i])
                revealed_acc += val * bproof.bl_hi[i]
        stmt = stmt & DLRep(self.bl_commit - revealed_acc, expr_bl_commit)

        # Having an invalid revealed secret leads to different
        # {commit-revealed_acc} value between prover and verifier. Since this
        # value is a zksk.statement constant, the mismatch results in a
        # StatementMismatch exception. verify_proof checks for this exception
        # and convert it to False
        try:
            is_valid = stmt.verify(bproof.nizk_proof)
        except StatementMismatch:
            return False
        return is_valid


def main():
    import doctest
    doctest.testmod(verbose=True)


if __name__ == "__main__":
    main()



__version__ = "0.2.1"
__title__ = "sscred"
__author__ = "Kasra EdalatNejad"
__email__ = "kasra.edalat@epfl.ch"
__url__ = "https://github.com/spring-epfl/SSCred"
__license__ = "BSD-3-Clause"
__description__ = "Single Show Credentials: A Python library for anonymous authentication."
__copyright__ = "2020, Kasra Edalatnejad  (SPRING Lab, EPFL)"

from sscred.commitment import *
from sscred.acl import *
from sscred.pack import *
from sscred.blind_pedersen import *
from sscred.blind_signature import *

